import shutil
import pandas as pd
from shapely.geometry import box
from utils.logger import setup_logger
from utils.gps_extractor import GPSExtractor
import numpy as np
import logging
from datetime import timedelta
import matplotlib.pyplot as plt
import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))


class TimeGroupOverlapFilter:
    """基于时间组重叠度的图像过滤器"""

    def __init__(self, image_dir: str, output_dir: str, overlap_threshold: float = 0.7):
        """
        初始化过滤器

        Args:
            image_dir: 图像目录
            output_dir: 输出目录
            overlap_threshold: 重叠阈值，默认0.7
        """
        self.image_dir = image_dir
        self.output_dir = output_dir
        self.overlap_threshold = overlap_threshold
        self.logger = logging.getLogger('UAV_Preprocess.TimeGroupFilter')

    def _group_by_time(self, points_df, time_threshold=timedelta(minutes=5)):
        """按时间间隔对点进行分组"""
        if 'date' not in points_df.columns:
            self.logger.error("数据中缺少date列")
            return []

        # 将date为空的行单独作为一组
        null_date_group = points_df[points_df['date'].isna()]
        valid_date_points = points_df[points_df['date'].notna()]

        if not null_date_group.empty:
            self.logger.info(f"发现 {len(null_date_group)} 个无时间戳的点，将作为单独分组")

        if valid_date_points.empty:
            self.logger.warning("没有有效的时间戳数据")
            return [null_date_group] if not null_date_group.empty else []

        # 按时间排序
        valid_date_points = valid_date_points.sort_values('date')

        # 计算时间差
        time_diffs = valid_date_points['date'].diff()

        # 找到时间差超过阈值的位置
        time_groups = []
        current_group_start = 0

        for idx, time_diff in enumerate(time_diffs):
            if time_diff and time_diff > time_threshold:
                # 添加当前组
                current_group = valid_date_points.iloc[current_group_start:idx]
                time_groups.append(current_group)
                current_group_start = idx

        # 添加最后一组
        last_group = valid_date_points.iloc[current_group_start:]
        if not last_group.empty:
            time_groups.append(last_group)

        # 如果有空时间戳的点，将其作为最后一组
        if not null_date_group.empty:
            time_groups.append(null_date_group)

        return time_groups

    def _get_group_bbox(self, group_df):
        """获取组内点的边界框"""
        min_lon = group_df['lon'].min()
        max_lon = group_df['lon'].max()
        min_lat = group_df['lat'].min()
        max_lat = group_df['lat'].max()
        return box(min_lon, min_lat, max_lon, max_lat)

    def _calculate_overlap(self, box1, box2):
        """计算两个边界框的重叠率"""
        if box1.intersects(box2):
            intersection_area = box1.intersection(box2).area
            smaller_area = min(box1.area, box2.area)
            return intersection_area / smaller_area
        return 0

    def filter_overlapping_groups(self, time_threshold=timedelta(minutes=5)):
        """过滤重叠的时间组"""
        # 提取GPS数据
        extractor = GPSExtractor(self.image_dir)
        gps_points = extractor.extract_all_gps()

        # 按时间分组
        time_groups = self._group_by_time(gps_points, time_threshold)

        # 计算每个组的边界框
        group_boxes = []
        for idx, group in enumerate(time_groups):
            if not group['date'].isna().any():  # 只处理有时间戳的组
                bbox = self._get_group_bbox(group)
                group_boxes.append((idx, group, bbox))

        # 找出需要删除的组
        groups_to_delete = set()
        for i in range(len(group_boxes)):
            if i in groups_to_delete:
                continue

            idx1, group1, box1 = group_boxes[i]
            area1 = box1.area

            for j in range(i + 1, len(group_boxes)):
                if j in groups_to_delete:
                    continue

                idx2, group2, box2 = group_boxes[j]
                area2 = box2.area

                overlap_ratio = self._calculate_overlap(box1, box2)

                if overlap_ratio > self.overlap_threshold:
                    # 删除面积较小的组
                    if area1 < area2:
                        group_to_delete = idx1
                        smaller_area = area1
                        larger_area = area2
                    else:
                        group_to_delete = idx2
                        smaller_area = area2
                        larger_area = area1

                    groups_to_delete.add(group_to_delete)
                    self.logger.info(
                        f"时间组 {group_to_delete + 1} 与时间组 "
                        f"{idx2 + 1 if group_to_delete == idx1 else idx1 + 1} "
                        f"重叠率为 {overlap_ratio:.2f}，"
                        f"面积比为 {smaller_area/larger_area:.2f}，"
                        f"将删除较小面积的组 {group_to_delete + 1}"
                    )

        # 删除重复组的图像
        deleted_files = []
        for group_idx in groups_to_delete:
            group_files = time_groups[group_idx]['file'].tolist()
            deleted_files.extend(group_files)

        self.logger.info(f"共删除 {len(groups_to_delete)} 个重复时间组，"
                         f"{len(deleted_files)} 张图像")

        # 可视化结果
        self._visualize_results(time_groups, groups_to_delete)

        return deleted_files

    def _visualize_results(self, time_groups, groups_to_delete):
        """可视化过滤结果"""
        plt.figure(figsize=(15, 10))

        # 生成不同的颜色
        colors = plt.cm.rainbow(np.linspace(0, 1, len(time_groups)))

        # 绘制所有组的边界框
        for idx, (group, color) in enumerate(zip(time_groups, colors)):
            if not group['date'].isna().any():  # 只处理有时间戳的组
                bbox = self._get_group_bbox(group)
                x, y = bbox.exterior.xy

                if idx in groups_to_delete:
                    # 被删除的组用虚线表示
                    plt.plot(x, y, '--', color=color, alpha=0.6,
                             label=f'Deleted Group {idx + 1}')
                else:
                    # 保留的组用实线表示
                    plt.plot(x, y, '-', color=color, alpha=0.6,
                             label=f'Group {idx + 1}')

                # 绘制该组的GPS点
                plt.scatter(group['lon'], group['lat'], color=color,
                            s=30, alpha=0.6)

        plt.title("Time Groups and Their Bounding Boxes", fontsize=14)
        plt.xlabel("Longitude", fontsize=12)
        plt.ylabel("Latitude", fontsize=12)
        plt.grid(True)
        plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left', fontsize=10)
        plt.tight_layout()

        # 保存图片
        plt.savefig(os.path.join(self.output_dir, 'filter_imgs', 'time_groups_overlap_bbox.png'),
                    dpi=300, bbox_inches='tight')
        plt.close()


if __name__ == '__main__':
    # 设置路径
    DATASET = r'F:\error_data\20241108134711\3D'
    output_dir = r'E:\studio2\ODM_pro\test'
    os.makedirs(output_dir, exist_ok=True)

    # 设置日志
    setup_logger(os.path.dirname(output_dir))

    # 创建过滤器并执行过滤
    filter = TimeGroupOverlapFilter(DATASET, output_dir, overlap_threshold=0.7)
    deleted_files = filter.filter_overlapping_groups(
        time_threshold=timedelta(minutes=5))
